@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013-2018 Joris Vink <joris@coders.se>
+ * Copyright (c) 2013-2021 Joris Vink <joris@coders.se>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -22,6 +22,7 @@
 #endif
 
 #include <sys/param.h>
+#include <sys/types.h>
 #include <sys/time.h>
 #include <sys/queue.h>
 #include <sys/un.h>
@@ -29,11 +30,9 @@
 #include <netinet/in.h>
 #include <arpa/inet.h>
 
-#if !defined(KORE_NO_TLS)
 #include <openssl/err.h>
 #include <openssl/dh.h>
 #include <openssl/ssl.h>
-#endif
 
 #include <errno.h>
 #include <regex.h>
@@ -56,16 +55,36 @@ extern int daemon(int, int);
 #define st_mtim		st_mtimespec
 #endif
 
-#if !defined(KORE_NO_SENDFILE) && defined(KORE_NO_TLS)
+#if !defined(KORE_NO_SENDFILE)
 #if defined(__MACH__) || defined(__FreeBSD_version) || defined(__linux__)
 #define KORE_USE_PLATFORM_SENDFILE	1
 #endif
 #endif
 
+/*
+ * Figure out what type of OpenSSL API we are dealing with.
+ */
+#if defined(LIBRESSL_VERSION_NUMBER)
+#if LIBRESSL_VERSION_NUMBER >= 0x3000000fL
+#define KORE_OPENSSL_NEWER_API		1
+#endif
+
+#if LIBRESSL_VERSION_NUMBER >= 0x3020200fL
+#define TLS1_3_VERSION			0x0304
+#endif
+
+#else
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+#define KORE_OPENSSL_NEWER_API		1
+#endif
+#endif
+
 #if defined(__OpenBSD__)
 #define KORE_USE_PLATFORM_PLEDGE	1
 #endif
 
+#define KORE_RSAKEY_BITS	4096
+
 #define KORE_RESULT_ERROR	0
 #define KORE_RESULT_OK		1
 #define KORE_RESULT_RETRY	2
@@ -74,6 +93,9 @@ extern int daemon(int, int);
 #define KORE_TLS_VERSION_1_2	1
 #define KORE_TLS_VERSION_BOTH	2
 
+#define KORE_BASE64_RAW		0x0001
+
+#define KORE_WAIT_INFINITE	(u_int64_t)-1
 #define KORE_RESEED_TIME	(1800 * 1000)
 
 #define errno_s			strerror(errno)
@@ -105,15 +127,15 @@ extern int daemon(int, int);
 #define NETBUF_IS_STREAM	0x10
 #define NETBUF_IS_FILEREF	0x20
 
-#define X509_GET_CN(c, o, l)					\
-	X509_NAME_get_text_by_NID(X509_get_subject_name(c),	\
-	    NID_commonName, o, l)
+#define KORE_X509_COMMON_NAME_ONLY	0x0001
 
-#define X509_CN_LENGTH		(ub_common_name + 1)
+#define KORE_PEM_CERT_CHAIN	1
+#define KORE_DER_CERT_DATA	2
 
 /* XXX hackish. */
 #if !defined(KORE_NO_HTTP)
 struct http_request;
+struct http_redirect;
 #endif
 
 #define KORE_FILEREF_SOFT_REMOVED	0x1000
@@ -121,16 +143,14 @@ struct http_request;
 struct kore_fileref {
 	int				cnt;
 	int				flags;
+	int				ontls;
 	off_t				size;
 	char				*path;
 	u_int64_t			mtime;
 	time_t				mtime_sec;
 	u_int64_t			expiration;
-#if !defined(KORE_USE_PLATFORM_SENDFILE)
 	void				*base;
-#else
 	int				fd;
-#endif
 	TAILQ_ENTRY(kore_fileref)	list;
 };
 
@@ -143,12 +163,10 @@ struct netbuf {
 	u_int8_t		flags;
 
 	struct kore_fileref	*file_ref;
-#if defined(KORE_USE_PLATFORM_SENDFILE)
 	off_t			fd_off;
 	off_t			fd_len;
-#endif
 
-	void			*owner;
+	struct connection	*owner;
 	void			*extra;
 	int			(*cb)(struct netbuf *);
 
@@ -162,6 +180,7 @@ TAILQ_HEAD(netbuf_head, netbuf);
 #define KORE_TYPE_PGSQL_CONN	3
 #define KORE_TYPE_TASK		4
 #define KORE_TYPE_PYSOCKET	5
+#define KORE_TYPE_CURL_HANDLE	6
 
 #define CONN_STATE_UNKNOWN		0
 #define CONN_STATE_TLS_SHAKE		1
@@ -172,14 +191,19 @@ TAILQ_HEAD(netbuf_head, netbuf);
 #define CONN_PROTO_HTTP		1
 #define CONN_PROTO_WEBSOCKET	2
 #define CONN_PROTO_MSG		3
+#define CONN_PROTO_ACME_ALPN	200
 
 #define KORE_EVENT_READ		0x01
 #define KORE_EVENT_WRITE	0x02
 #define KORE_EVENT_ERROR	0x04
 
-#define CONN_IDLE_TIMER_ACT	0x01
-#define CONN_CLOSE_EMPTY	0x02
-#define CONN_WS_CLOSE_SENT	0x04
+#define CONN_IDLE_TIMER_ACT	0x0001
+#define CONN_CLOSE_EMPTY	0x0002
+#define CONN_WS_CLOSE_SENT	0x0004
+#define CONN_IS_BUSY		0x0008
+#define CONN_LOG_TLS_FAILURE	0x0020
+#define CONN_TLS_ALPN_ACME_SEEN	0x0040
+#define CONN_TLS_SNI_SEEN	0x0080
 
 #define KORE_IDLE_TIMER_MAX	5000
 
@@ -210,13 +234,12 @@ struct connection {
 	int			fd;
 	u_int8_t		state;
 	u_int8_t		proto;
-	void			*owner;
-#if !defined(KORE_NO_TLS)
+	struct listener		*owner;
 	X509			*cert;
 	SSL			*ssl;
+	char			*tls_sni;
 	int			tls_reneg;
-#endif
-	u_int8_t		flags;
+	u_int16_t		flags;
 	void			*hdlr_extra;
 
 	int			(*handle)(struct connection *);
@@ -241,6 +264,8 @@ struct connection {
 	struct netbuf		*rnb;
 
 #if !defined(KORE_NO_HTTP)
+	u_int64_t			http_start;
+	u_int64_t			http_timeout;
 	struct kore_runtime_call	*ws_connect;
 	struct kore_runtime_call	*ws_message;
 	struct kore_runtime_call	*ws_disconnect;
@@ -278,18 +303,60 @@ struct kore_runtime_call {
 	struct kore_runtime	*runtime;
 };
 
+struct kore_domain {
+	u_int16_t				id;
+	int					logerr;
+	u_int64_t				logwarn;
+	int					accesslog;
+
+	char					*domain;
+	struct kore_buf				*logbuf;
+	struct kore_server			*server;
+
+#if defined(KORE_USE_ACME)
+	int					acme;
+	int					acme_challenge;
+	void					*acme_cert;
+	size_t					acme_cert_len;
+#endif
+	char					*cafile;
+	char					*crlfile;
+	char					*certfile;
+	char					*certkey;
+	SSL_CTX					*ssl_ctx;
+	int					x509_verify_depth;
+#if !defined(KORE_NO_HTTP)
+	TAILQ_HEAD(, kore_module_handle)	handlers;
+	TAILQ_HEAD(, http_redirect)		redirects;
+#endif
+	TAILQ_ENTRY(kore_domain)		list;
+};
+
+TAILQ_HEAD(kore_domain_h, kore_domain);
+
 extern struct kore_runtime	kore_native_runtime;
 
 struct listener {
 	struct kore_event		evt;
 	int				fd;
 	int				family;
+	char				*port;
+	char				*host;
+	struct kore_server		*server;
 	struct kore_runtime_call	*connect;
 
-	LIST_ENTRY(listener)	list;
+	LIST_ENTRY(listener)		list;
 };
 
-LIST_HEAD(listener_head, listener);
+struct kore_server {
+	int				tls;
+	char				*name;
+	struct kore_domain_h		domains;
+	LIST_HEAD(, listener)		listeners;
+	LIST_ENTRY(kore_server)		list;
+};
+
+LIST_HEAD(kore_server_list, kore_server);
 
 #if !defined(KORE_NO_HTTP)
 
@@ -344,7 +411,6 @@ struct kore_module {
 	char				*path;
 	char				*onload;
 	int				type;
-	time_t				mtime;
 	struct kore_runtime_call	*ocb;
 
 	struct kore_module_functions	*fun;
@@ -354,6 +420,7 @@ struct kore_module {
 };
 
 #if !defined(KORE_NO_HTTP)
+
 struct kore_module_handle {
 	char					*path;
 	char					*func;
@@ -383,8 +450,12 @@ struct kore_alog_header {
 } __attribute__((packed));
 
 struct kore_worker {
-	u_int8_t			id;
-	u_int8_t			cpu;
+	u_int16_t			id;
+	u_int16_t			cpu;
+	int				running;
+#if defined(__linux__)
+	int				tracing;
+#endif
 	pid_t				pid;
 	int				pipe[2];
 	struct connection		*msg[2];
@@ -401,29 +472,6 @@ struct kore_worker {
 	} lb;
 };
 
-struct kore_domain {
-	u_int16_t				id;
-	char					*domain;
-	int					accesslog;
-	struct kore_buf				*logbuf;
-	int					logerr;
-	u_int64_t				logwarn;
-#if !defined(KORE_NO_TLS)
-	char					*cafile;
-	char					*crlfile;
-	char					*certfile;
-	char					*certkey;
-	SSL_CTX					*ssl_ctx;
-	int					x509_verify_depth;
-#endif
-#if !defined(KORE_NO_HTTP)
-	TAILQ_HEAD(, kore_module_handle)	handlers;
-#endif
-	TAILQ_ENTRY(kore_domain)		list;
-};
-
-TAILQ_HEAD(kore_domain_h, kore_domain);
-
 #if !defined(KORE_NO_HTTP)
 
 #define KORE_VALIDATOR_TYPE_REGEX	1
@@ -449,6 +497,107 @@ struct kore_buf {
 	size_t			offset;
 };
 
+#define KORE_JSON_TYPE_OBJECT		0x0001
+#define KORE_JSON_TYPE_ARRAY		0x0002
+#define KORE_JSON_TYPE_STRING		0x0004
+#define KORE_JSON_TYPE_NUMBER		0x0008
+#define KORE_JSON_TYPE_LITERAL		0x0010
+#define KORE_JSON_TYPE_INTEGER		0x0020
+#define KORE_JSON_TYPE_INTEGER_U64	0x0040
+
+#define KORE_JSON_FALSE			0
+#define KORE_JSON_TRUE			1
+#define KORE_JSON_NULL			2
+
+#define KORE_JSON_DEPTH_MAX		10
+
+#define KORE_JSON_ERR_NONE		0
+#define KORE_JSON_ERR_INVALID_OBJECT	1
+#define KORE_JSON_ERR_INVALID_ARRAY	2
+#define KORE_JSON_ERR_INVALID_STRING	3
+#define KORE_JSON_ERR_INVALID_NUMBER	4
+#define KORE_JSON_ERR_INVALID_LITERAL	5
+#define KORE_JSON_ERR_DEPTH		6
+#define KORE_JSON_ERR_EOF		7
+#define KORE_JSON_ERR_INVALID_JSON	8
+#define KORE_JSON_ERR_INVALID_SEARCH	9
+#define KORE_JSON_ERR_NOT_FOUND		10
+#define KORE_JSON_ERR_TYPE_MISMATCH	11
+#define KORE_JSON_ERR_LAST		KORE_JSON_ERR_TYPE_MISMATCH
+
+#define kore_json_find_object(j, p)		\
+    kore_json_find(j, p, KORE_JSON_TYPE_OBJECT)
+
+#define kore_json_find_array(j, p)		\
+    kore_json_find(j, p, KORE_JSON_TYPE_ARRAY)
+
+#define kore_json_find_string(j, p)		\
+    kore_json_find(j, p, KORE_JSON_TYPE_STRING)
+
+#define kore_json_find_number(j, p)		\
+    kore_json_find(j, p, KORE_JSON_TYPE_NUMBER)
+
+#define kore_json_find_integer(j, p)		\
+    kore_json_find(j, p, KORE_JSON_TYPE_INTEGER)
+
+#define kore_json_find_integer_u64(j, p)	\
+    kore_json_find(j, p, KORE_JSON_TYPE_INTEGER_U64)
+
+#define kore_json_find_literal(j, p)		\
+    kore_json_find(j, p, KORE_JSON_TYPE_LITERAL)
+
+#define kore_json_create_object(o, n)				\
+    kore_json_create_item(o, n, KORE_JSON_TYPE_OBJECT)
+
+#define kore_json_create_array(o, n)				\
+    kore_json_create_item(o, n, KORE_JSON_TYPE_ARRAY)
+
+#define kore_json_create_string(o, n, v)			\
+    kore_json_create_item(o, n, KORE_JSON_TYPE_STRING, v)
+
+#define kore_json_create_number(o, n, v)			\
+    kore_json_create_item(o, n, KORE_JSON_TYPE_NUMBER, v)
+
+#define kore_json_create_integer(o, n, v)			\
+    kore_json_create_item(o, n, KORE_JSON_TYPE_INTEGER, v)
+
+#define kore_json_create_integer_u64(o, n, v)			\
+    kore_json_create_item(o, n, KORE_JSON_TYPE_INTEGER_U64, v)
+
+#define kore_json_create_literal(o, n, v)			\
+    kore_json_create_item(o, n, KORE_JSON_TYPE_LITERAL, v)
+
+struct kore_json {
+	const u_int8_t			*data;
+	int				depth;
+	int				error;
+	size_t				length;
+	size_t				offset;
+
+	struct kore_buf			tmpbuf;
+	struct kore_json_item		*root;
+};
+
+struct kore_json_item {
+	u_int32_t			type;
+	char				*name;
+	struct kore_json_item		*parent;
+
+	union {
+		TAILQ_HEAD(, kore_json_item)	items;
+		char				*string;
+		double				number;
+		int				literal;
+		int64_t				s64;
+		u_int64_t			u64;
+	} data;
+
+	int				(*parse)(struct kore_json *,
+					    struct kore_json_item *);
+
+	TAILQ_ENTRY(kore_json_item)	list;
+};
+
 struct kore_pool_region {
 	void				*start;
 	size_t				length;
@@ -484,7 +633,19 @@ struct kore_timer {
 	TAILQ_ENTRY(kore_timer)	list;
 };
 
-#define KORE_WORKER_KEYMGR	0
+/*
+ * Keymgr process is worker index 0, but id 2000.
+ * Acme process is worker index 1, but id 2001.
+ */
+#define KORE_WORKER_KEYMGR_IDX		0
+#define KORE_WORKER_ACME_IDX		1
+#define KORE_WORKER_BASE		2
+#define KORE_WORKER_KEYMGR		2000
+#define KORE_WORKER_ACME		2001
+#define KORE_WORKER_MAX			UCHAR_MAX
+
+#define KORE_WORKER_POLICY_RESTART	1
+#define KORE_WORKER_POLICY_TERMINATE	2
 
 /* Reserved message ids, registered on workers. */
 #define KORE_MSG_WEBSOCKET		1
@@ -496,6 +657,12 @@ struct kore_timer {
 #define KORE_MSG_CERTIFICATE		7
 #define KORE_MSG_CERTIFICATE_REQ	8
 #define KORE_MSG_CRL			9
+#define KORE_MSG_ACCEPT_AVAILABLE	10
+#define KORE_PYTHON_SEND_OBJ		11
+#define KORE_MSG_ACME_BASE		100
+
+/* messages for applications should start at 201. */
+#define KORE_MSG_APP_BASE		200
 
 /* Predefined message targets. */
 #define KORE_MSG_PARENT		1000
@@ -508,33 +675,30 @@ struct kore_msg {
 	size_t		length;
 };
 
-#if !defined(KORE_NO_TLS)
 struct kore_keyreq {
 	int		padding;
-	char		domain[KORE_DOMAINNAME_LEN];
-	u_int16_t	domain_len;
-	u_int16_t	data_len;
+	char		domain[KORE_DOMAINNAME_LEN + 1];
+	size_t		data_len;
 	u_int8_t	data[];
 };
 
 struct kore_x509_msg {
-	char		domain[KORE_DOMAINNAME_LEN];
-	u_int16_t	domain_len;
+	char		domain[KORE_DOMAINNAME_LEN + 1];
 	size_t		data_len;
 	u_int8_t	data[];
 };
-#endif
 
 #if !defined(KORE_SINGLE_BINARY)
 extern char	*config_file;
 #endif
 
 extern pid_t	kore_pid;
-extern int	foreground;
 extern int	kore_quiet;
 extern int	kore_debug;
 extern int	skip_chroot;
 extern int	skip_runas;
+extern int	kore_foreground;
+
 extern char	*kore_pidfile;
 extern char	*kore_root_path;
 extern char	*kore_runas_user;
@@ -542,18 +706,20 @@ extern char	*kore_tls_cipher_list;
 
 extern volatile sig_atomic_t	sig_recv;
 
-#if !defined(KORE_NO_TLS)
 extern int	tls_version;
 extern DH	*tls_dhparam;
 extern char	*rand_file;
+extern int	keymgr_active;
 extern char	*keymgr_runas_user;
 extern char	*keymgr_root_path;
-#endif
+extern char	*acme_runas_user;
+extern char	*acme_root_path;
 
 extern u_int8_t			nlisteners;
 extern u_int16_t		cpu_count;
 extern u_int8_t			worker_count;
 extern const char		*kore_version;
+extern int			worker_policy;
 extern u_int8_t			worker_set_affinity;
 extern u_int32_t		worker_rlimit_nofiles;
 extern u_int32_t		worker_max_connections;
@@ -563,36 +729,44 @@ extern u_int64_t		kore_websocket_maxframe;
 extern u_int64_t		kore_websocket_timeout;
 extern u_int32_t		kore_socket_backlog;
 
-extern struct listener_head	listeners;
 extern struct kore_worker	*worker;
-extern struct kore_domain_h	domains;
-extern struct kore_domain	*primary_dom;
 extern struct kore_pool		nb_pool;
+extern struct kore_domain	*primary_dom;
+extern struct kore_server_list	kore_servers;
 
 void		kore_signal(int);
 void		kore_shutdown(void);
 void		kore_signal_setup(void);
-void		kore_worker_wait(int);
+void		kore_proctitle(const char *);
+void		kore_default_getopt(int, char **);
+
+void		kore_worker_reap(void);
 void		kore_worker_init(void);
 void		kore_worker_make_busy(void);
 void		kore_worker_shutdown(void);
 void		kore_worker_dispatch_signal(int);
-void		kore_worker_spawn(u_int16_t, u_int16_t);
-void		kore_worker_entry(struct kore_worker *);
 void		kore_worker_privdrop(const char *, const char *);
+void		kore_worker_spawn(u_int16_t, u_int16_t, u_int16_t);
+int		kore_worker_keymgr_response_verify(struct kore_msg *,
+		    const void *, struct kore_domain **);
+
+void	kore_worker_entry(struct kore_worker *) __attribute__((noreturn));
 
 struct kore_worker	*kore_worker_data(u_int8_t);
 
 void		kore_platform_init(void);
+void		kore_platform_sandbox(void);
 void		kore_platform_event_init(void);
 void		kore_platform_event_cleanup(void);
-void		kore_platform_proctitle(char *);
 void		kore_platform_disable_read(int);
 void		kore_platform_disable_write(int);
 void		kore_platform_enable_accept(void);
 void		kore_platform_disable_accept(void);
 void		kore_platform_event_wait(u_int64_t);
 void		kore_platform_event_all(int, void *);
+void		kore_platform_event_level_all(int, void *);
+void		kore_platform_event_level_read(int, void *);
+void		kore_platform_proctitle(const char *);
 void		kore_platform_schedule_read(int, void *);
 void		kore_platform_schedule_write(int, void *);
 void		kore_platform_event_schedule(int, int, int, void *);
@@ -623,23 +797,34 @@ struct kore_auth	*kore_auth_lookup(const char *);
 #endif
 
 void		kore_timer_init(void);
-u_int64_t	kore_timer_run(u_int64_t);
+void		kore_timer_run(u_int64_t);
+u_int64_t	kore_timer_next_run(u_int64_t);
 void		kore_timer_remove(struct kore_timer *);
 struct kore_timer	*kore_timer_add(void (*cb)(void *, u_int64_t),
 			    u_int64_t, void *, int);
 
-void		kore_listener_cleanup(void);
+void		kore_server_closeall(void);
+void		kore_server_cleanup(void);
+void		kore_server_free(struct kore_server *);
+void		kore_server_finalize(struct kore_server *);
+
+struct kore_server	*kore_server_create(const char *);
+struct kore_server	*kore_server_lookup(const char *);
+
 void		kore_listener_accept(void *, int);
+struct listener	*kore_listener_lookup(const char *);
 void		kore_listener_free(struct listener *);
-struct listener	*kore_listener_alloc(int, const char *);
+struct listener	*kore_listener_create(struct kore_server *);
+int		kore_listener_init(struct listener *, int, const char *);
 
 int		kore_sockopt(int, int, int);
-int		kore_server_bind_unix(const char *, const char *);
-int		kore_server_bind(const char *, const char *, const char *);
-#if !defined(KORE_NO_TLS)
+int		kore_server_bind_unix(struct kore_server *,
+		    const char *, const char *);
+int		kore_server_bind(struct kore_server *,
+		    const char *, const char *, const char *);
+
 int		kore_tls_sni_cb(SSL *, int *, void *);
 void		kore_tls_info_callback(const SSL *, int, int);
-#endif
 
 void			kore_connection_init(void);
 void			kore_connection_cleanup(void);
@@ -661,6 +846,10 @@ int			kore_connection_accept(struct listener *,
 u_int64_t	kore_time_ms(void);
 void		kore_log_init(void);
 
+#if defined(KORE_USE_PYTHON)
+int		kore_configure_setting(const char *, char *);
+#endif
+
 void		*kore_malloc(size_t);
 void		kore_parse_config(void);
 void		kore_parse_config_file(FILE *);
@@ -695,10 +884,16 @@ long long	kore_strtonum(const char *, int, long long, long long, int *);
 double		kore_strtodouble(const char *, long double, long double, int *);
 int		kore_base64_encode(const void *, size_t, char **);
 int		kore_base64_decode(const char *, u_int8_t **, size_t *);
+int		kore_base64url_encode(const void *, size_t, char **, int);
+int		kore_base64url_decode(const char *, u_int8_t **, size_t *, int);
 void		*kore_mem_find(void *, size_t, const void *, size_t);
 char		*kore_text_trim(char *, size_t);
 char		*kore_read_line(FILE *, char *, size_t);
 
+EVP_PKEY	*kore_rsakey_load(const char *);
+EVP_PKEY	*kore_rsakey_generate(const char *);
+int		kore_x509_subject_name(struct connection *, char **, int);
+
 #if !defined(KORE_NO_HTTP)
 void		kore_websocket_handshake(struct http_request *,
 		    const char *, const char *, const char *);
@@ -712,6 +907,7 @@ void		kore_websocket_broadcast(struct connection *,
 void		kore_msg_init(void);
 void		kore_msg_worker_init(void);
 void		kore_msg_parent_init(void);
+void		kore_msg_unregister(u_int8_t);
 void		kore_msg_parent_add(struct kore_worker *);
 void		kore_msg_parent_remove(struct kore_worker *);
 void		kore_msg_send(u_int16_t, u_int8_t, const void *, size_t);
@@ -728,14 +924,15 @@ extern char	*kore_filemap_index;
 #endif
 
 void			kore_fileref_init(void);
-struct kore_fileref	*kore_fileref_get(const char *);
-struct kore_fileref	*kore_fileref_create(const char *, int, off_t,
-			    struct timespec *);
+struct kore_fileref	*kore_fileref_get(const char *, int);
+struct kore_fileref	*kore_fileref_create(struct kore_server *,
+			    const char *, int, off_t, struct timespec *);
 void			kore_fileref_release(struct kore_fileref *);
 
+struct kore_domain	*kore_domain_new(const char *);
+
 void		kore_domain_init(void);
 void		kore_domain_cleanup(void);
-int		kore_domain_new(char *);
 void		kore_domain_free(struct kore_domain *);
 void		kore_module_init(void);
 void		kore_module_cleanup(void);
@@ -747,14 +944,16 @@ void		*kore_module_getsym(const char *, struct kore_runtime **);
 void		kore_domain_load_crl(void);
 void		kore_domain_keymgr_init(void);
 void		kore_domain_callback(void (*cb)(struct kore_domain *));
-void		kore_domain_tlsinit(struct kore_domain *, const void *, size_t);
+int		kore_domain_attach(struct kore_domain *, struct kore_server *);
+void		kore_domain_tlsinit(struct kore_domain *, int,
+		    const void *, size_t);
 void		kore_domain_crl_add(struct kore_domain *, const void *, size_t);
 #if !defined(KORE_NO_HTTP)
-int		kore_module_handler_new(const char *, const char *,
+int		kore_module_handler_new(struct kore_domain *, const char *,
 		    const char *, const char *, int);
 void		kore_module_handler_free(struct kore_module_handle *);
-struct kore_module_handle	*kore_module_handler_find(const char *,
-				    const char *);
+struct kore_module_handle	*kore_module_handler_find(struct http_request *,
+				    struct kore_domain *);
 #endif
 
 struct kore_runtime_call	*kore_runtime_getcall(const char *);
@@ -777,8 +976,8 @@ void	kore_runtime_wsmessage(struct kore_runtime_call *,
 	    struct connection *, u_int8_t, const void *, size_t);
 #endif
 
-struct kore_domain		*kore_domain_byid(u_int16_t);
-struct kore_domain		*kore_domain_lookup(const char *);
+struct kore_domain	*kore_domain_byid(u_int16_t);
+struct kore_domain	*kore_domain_lookup(struct kore_server *, const char *);
 
 #if !defined(KORE_NO_HTTP)
 void		kore_validator_init(void);
@@ -792,6 +991,8 @@ struct kore_validator	*kore_validator_lookup(const char *);
 
 void		fatal(const char *, ...) __attribute__((noreturn));
 void		fatalx(const char *, ...) __attribute__((noreturn));
+
+const char	*kore_worker_name(int);
 void		kore_debug_internal(char *, int, const char *, ...);
 
 u_int16_t	net_read16(u_int8_t *);
@@ -837,15 +1038,21 @@ void	kore_buf_appendv(struct kore_buf *, const char *, va_list);
 void	kore_buf_replace_string(struct kore_buf *,
 	    const char *, const void *, size_t);
 
+int	kore_json_parse(struct kore_json *);
+void	kore_json_cleanup(struct kore_json *);
+void	kore_json_item_free(struct kore_json_item *);
+void	kore_json_init(struct kore_json *, const void *, size_t);
+void	kore_json_item_tobuf(struct kore_json_item *, struct kore_buf *);
+
+const char		*kore_json_strerror(struct kore_json *);
+struct kore_json_item	*kore_json_find(struct kore_json_item *,
+			    const char *, u_int32_t);
+struct kore_json_item	*kore_json_create_item(struct kore_json_item *,
+			    const char *, u_int32_t, ...);
+
 void	kore_keymgr_run(void);
 void	kore_keymgr_cleanup(int);
 
-void	kore_worker_teardown(void);
-void	kore_parent_teardown(void);
-void	kore_worker_configure(void);
-void	kore_parent_daemonized(void);
-void	kore_parent_configure(int, char **);
-
 #if defined(__cplusplus)
 }
 #endif
